@@ -30,7 +30,6 @@
 #include "gstvaapivideocontext.h"
 #include "gstvaapivideometa.h"
 #include "gstvaapivideobufferpool.h"
-#include <gst/allocators/allocators.h>
 
 /* Default debug category is from the subclass */
 #define GST_CAT_DEFAULT (plugin->debug_category)
@@ -92,6 +91,21 @@ default_display_changed (GstVaapiPluginBase * plugin)
 {
 }
 
+static GstVaapiSurface *
+_get_cached_surface (GstBuffer * buf)
+{
+  return gst_mini_object_get_qdata (GST_MINI_OBJECT (buf),
+      g_quark_from_static_string ("GstVaapiDMABufSurface"));
+}
+
+static void
+_set_cached_surface (GstBuffer * buf, GstVaapiSurface * surface)
+{
+  return gst_mini_object_set_qdata (GST_MINI_OBJECT (buf),
+      g_quark_from_static_string ("GstVaapiDMABufSurface"), surface,
+      (GDestroyNotify) gst_vaapi_object_unref);
+}
+
 static gboolean
 plugin_update_sinkpad_info_from_buffer (GstVaapiPluginBase * plugin,
     GstBuffer * buf)
@@ -152,34 +166,62 @@ plugin_bind_dma_to_vaapi_buffer (GstVaapiPluginBase * plugin,
   meta = gst_buffer_get_vaapi_video_meta (outbuf);
   g_return_val_if_fail (meta != NULL, FALSE);
 
-  surface = gst_vaapi_surface_new_with_dma_buf_handle (plugin->display, fd,
-      GST_VIDEO_INFO_SIZE (vip), GST_VIDEO_INFO_FORMAT (vip),
-      GST_VIDEO_INFO_WIDTH (vip), GST_VIDEO_INFO_HEIGHT (vip),
-      vip->offset, vip->stride);
-  if (!surface)
-    goto error_create_surface;
+  /* Check for a VASurface cached in the buffer */
+  surface = _get_cached_surface (inbuf);
+  if (!surface) {
+    /* otherwise create one and cache it */
+    surface =
+        gst_vaapi_surface_new_with_dma_buf_handle (plugin->display, fd, vip);
+    if (!surface)
+      goto error_create_surface;
+    _set_cached_surface (inbuf, surface);
+  }
 
   proxy = gst_vaapi_surface_proxy_new (surface);
-  gst_vaapi_object_unref (surface);
   if (!proxy)
     goto error_create_proxy;
-
-  gst_vaapi_surface_proxy_set_destroy_notify (proxy,
-      (GDestroyNotify) gst_buffer_unref, (gpointer) gst_buffer_ref (inbuf));
   gst_vaapi_video_meta_set_surface_proxy (meta, proxy);
   gst_vaapi_surface_proxy_unref (proxy);
+  gst_buffer_add_parent_buffer_meta (outbuf, inbuf);
   return TRUE;
 
   /* ERRORS */
 error_update_sinkpad_info:
-  GST_ERROR ("failed to update sink pad video info from video meta");
-  return FALSE;
+  {
+    GST_ERROR_OBJECT (plugin,
+        "failed to update sink pad video info from video meta");
+    return FALSE;
+  }
 error_create_surface:
-  GST_ERROR ("failed to create VA surface from dma_buf handle");
-  return FALSE;
+  {
+    GST_ERROR_OBJECT (plugin,
+        "failed to create VA surface from dma_buf handle");
+    return FALSE;
+  }
 error_create_proxy:
-  GST_ERROR ("failed to create VA surface proxy from wrapped VA surface");
-  return FALSE;
+  {
+    GST_ERROR_OBJECT (plugin,
+        "failed to create VA surface proxy from wrapped VA surface");
+    return FALSE;
+  }
+}
+
+static void
+plugin_reset_texture_map (GstVaapiPluginBase * plugin)
+{
+  if (plugin->display)
+    gst_vaapi_display_reset_texture_map (plugin->display);
+}
+
+static void
+gst_vaapi_plugin_base_find_gl_context (GstVaapiPluginBase * plugin)
+{
+  GstObject *gl_context;
+
+  if (!gst_vaapi_find_gl_local_context (GST_ELEMENT_CAST (plugin), &gl_context))
+    return;
+  gst_vaapi_plugin_base_set_gl_context (plugin, gl_context);
+  gst_object_unref (gl_context);
 }
 
 void
@@ -243,11 +285,13 @@ gst_vaapi_plugin_base_open (GstVaapiPluginBase * plugin)
 void
 gst_vaapi_plugin_base_close (GstVaapiPluginBase * plugin)
 {
+  /* Release vaapi textures first if exist, which refs display object */
+  plugin_reset_texture_map (plugin);
+
   gst_vaapi_display_replace (&plugin->display, NULL);
   gst_object_replace (&plugin->gl_context, NULL);
 
   gst_caps_replace (&plugin->sinkpad_caps, NULL);
-  plugin->sinkpad_caps_changed = FALSE;
   gst_video_info_init (&plugin->sinkpad_info);
   if (plugin->sinkpad_buffer_pool) {
     gst_object_unref (plugin->sinkpad_buffer_pool);
@@ -255,8 +299,10 @@ gst_vaapi_plugin_base_close (GstVaapiPluginBase * plugin)
   }
   g_clear_object (&plugin->srcpad_buffer_pool);
 
+  g_clear_object (&plugin->sinkpad_allocator);
+  g_clear_object (&plugin->srcpad_allocator);
+
   gst_caps_replace (&plugin->srcpad_caps, NULL);
-  plugin->srcpad_caps_changed = FALSE;
   gst_video_info_init (&plugin->srcpad_info);
   gst_caps_replace (&plugin->allowed_raw_caps, NULL);
 }
@@ -338,6 +384,10 @@ gst_vaapi_plugin_base_ensure_display (GstVaapiPluginBase * plugin)
     return TRUE;
   gst_vaapi_display_replace (&plugin->display, NULL);
 
+  /* Query for a local GstGL context. If it's found, it will be used
+   * to create the VA display */
+  gst_vaapi_plugin_base_find_gl_context (plugin);
+
   if (!gst_vaapi_ensure_display (GST_ELEMENT (plugin),
           plugin->display_type_req))
     return FALSE;
@@ -409,6 +459,167 @@ has_dmabuf_capable_peer (GstVaapiPluginBase * plugin, GstPad * pad)
   return is_dmabuf_capable;
 }
 
+static gboolean
+set_dmabuf_allocator (GstVaapiPluginBase * plugin, GstBufferPool * pool,
+    GstCaps * caps)
+{
+  GstStructure *config;
+  GstAllocator *allocator;
+  GstVideoInfo vi;
+
+  if (!gst_video_info_from_caps (&vi, caps))
+    return FALSE;
+  allocator = gst_vaapi_dmabuf_allocator_new (plugin->display, &vi,
+      GST_VAAPI_SURFACE_ALLOC_FLAG_LINEAR_STORAGE);
+  if (!allocator)
+    return FALSE;
+
+  config = gst_buffer_pool_get_config (pool);
+  gst_buffer_pool_config_set_allocator (config, allocator, NULL);
+  if (!gst_buffer_pool_set_config (pool, config))
+    return FALSE;
+
+  if (plugin->sinkpad_allocator)
+    gst_object_unref (plugin->sinkpad_allocator);
+  plugin->sinkpad_allocator = allocator;
+  return TRUE;
+}
+
+static gboolean
+gst_vaapi_buffer_pool_caps_is_equal (GstBufferPool * pool, GstCaps * newcaps)
+{
+  GstStructure *config;
+  GstCaps *caps;
+  gboolean ret;
+
+  caps = NULL;
+  ret = FALSE;
+  config = gst_buffer_pool_get_config (pool);
+  if (gst_buffer_pool_config_get_params (config, &caps, NULL, NULL, NULL))
+    ret = gst_caps_is_equal (newcaps, caps);
+  gst_structure_free (config);
+
+  return ret;
+}
+
+static inline gboolean
+reset_allocator (GstAllocator * allocator, GstVideoInfo * vinfo)
+{
+  const GstVideoInfo *orig_vi;
+
+  if (!allocator)
+    return TRUE;
+
+  orig_vi = gst_allocator_get_vaapi_video_info (allocator, NULL);
+  if (!gst_video_info_changed (orig_vi, vinfo))
+    return FALSE;
+
+  gst_object_unref (allocator);
+  return TRUE;
+}
+
+static gboolean
+ensure_sinkpad_allocator (GstVaapiPluginBase * plugin, GstVideoInfo * vinfo)
+{
+  if (!reset_allocator (plugin->sinkpad_allocator, vinfo))
+    return TRUE;
+
+  if (has_dmabuf_capable_peer (plugin, plugin->sinkpad)) {
+    plugin->sinkpad_allocator =
+        gst_vaapi_dmabuf_allocator_new (plugin->display, vinfo,
+        GST_VAAPI_SURFACE_ALLOC_FLAG_LINEAR_STORAGE);
+  } else {
+    plugin->sinkpad_allocator =
+        gst_vaapi_video_allocator_new (plugin->display, vinfo, 0);
+  }
+  return plugin->sinkpad_allocator != NULL;
+}
+
+static gboolean
+ensure_srcpad_allocator (GstVaapiPluginBase * plugin, GstVideoInfo * vinfo)
+{
+  if (!reset_allocator (plugin->srcpad_allocator, vinfo))
+    return TRUE;
+
+  plugin->srcpad_allocator =
+      gst_vaapi_video_allocator_new (plugin->display, vinfo, 0);
+  return plugin->srcpad_allocator != NULL;
+}
+
+/**
+ * gst_vaapi_plugin_base_create_pool:
+ * @plugin: a #GstVaapiPluginBase
+ * @caps: the initial #GstCaps for the resulting buffer pool
+ * @size: the size of each buffer, not including prefix and padding
+ * @options: a set of #GstVaapiVideoBufferPoolOption encoded as bit-wise
+ * @allocator: (allow-none): the #GstAllocator to use or %NULL
+ *
+ * Create an instance of #GstVaapiVideoBufferPool
+ *
+ * Returns: (transfer full): a new allocated #GstBufferPool
+ **/
+static GstBufferPool *
+gst_vaapi_plugin_base_create_pool (GstVaapiPluginBase * plugin, GstCaps * caps,
+    gsize size, guint min_buffers, guint max_buffers, guint options,
+    GstAllocator * allocator)
+{
+  GstBufferPool *pool;
+  GstStructure *config;
+
+  if (!(pool = gst_vaapi_video_buffer_pool_new (plugin->display)))
+    goto error_create_pool;
+
+  config = gst_buffer_pool_get_config (pool);
+  gst_buffer_pool_config_set_params (config, caps, size, min_buffers,
+      max_buffers);
+  gst_buffer_pool_config_add_option (config,
+      GST_BUFFER_POOL_OPTION_VAAPI_VIDEO_META);
+  if (options & GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META) {
+    gst_buffer_pool_config_add_option (config,
+        GST_BUFFER_POOL_OPTION_VIDEO_META);
+  }
+  if (options & GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT) {
+    gst_buffer_pool_config_add_option (config,
+        GST_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT);
+  }
+#if (USE_GLX || USE_EGL)
+  if (options & GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_GL_TEXTURE_UPLOAD) {
+    gst_buffer_pool_config_add_option (config,
+        GST_BUFFER_POOL_OPTION_VIDEO_GL_TEXTURE_UPLOAD_META);
+  }
+#endif
+  if (allocator)
+    gst_buffer_pool_config_set_allocator (config, allocator, NULL);
+  if (!gst_buffer_pool_set_config (pool, config)) {
+    config = gst_buffer_pool_get_config (pool);
+
+    if (!gst_buffer_pool_config_validate_params (config, caps, size,
+            min_buffers, max_buffers)) {
+      gst_structure_free (config);
+      goto error_pool_config;
+    }
+
+    if (!gst_buffer_pool_set_config (pool, config))
+      goto error_pool_config;
+  }
+  return pool;
+
+  /* ERRORS */
+error_create_pool:
+  {
+    GST_ERROR_OBJECT (plugin, "failed to create buffer pool");
+    return NULL;
+  }
+error_pool_config:
+  {
+    gst_object_unref (pool);
+    GST_ELEMENT_ERROR (plugin, RESOURCE, SETTINGS,
+        ("Failed to configure the buffer pool"),
+        ("Configuration is most likely invalid, please report this issue."));
+    return NULL;
+  }
+}
+
 /**
  * ensure_sinkpad_buffer_pool:
  * @plugin: a #GstVaapiPluginBase
@@ -423,59 +634,56 @@ static gboolean
 ensure_sinkpad_buffer_pool (GstVaapiPluginBase * plugin, GstCaps * caps)
 {
   GstBufferPool *pool;
-  GstCaps *pool_caps;
-  GstStructure *config;
   GstVideoInfo vi;
-  gboolean need_pool;
+  guint size;
+
+  /* video decoders don't use a buffer pool in the sink pad */
+  if (GST_IS_VIDEO_DECODER (plugin))
+    return TRUE;
 
   if (!gst_vaapi_plugin_base_ensure_display (plugin))
     return FALSE;
 
   if (plugin->sinkpad_buffer_pool) {
-    config = gst_buffer_pool_get_config (plugin->sinkpad_buffer_pool);
-    gst_buffer_pool_config_get_params (config, &pool_caps, NULL, NULL, NULL);
-    need_pool = !gst_caps_is_equal (caps, pool_caps);
-    gst_structure_free (config);
-    if (!need_pool)
+    if (gst_vaapi_buffer_pool_caps_is_equal (plugin->sinkpad_buffer_pool, caps))
       return TRUE;
+    gst_buffer_pool_set_active (plugin->sinkpad_buffer_pool, FALSE);
     g_clear_object (&plugin->sinkpad_buffer_pool);
+    g_clear_object (&plugin->sinkpad_allocator);
     plugin->sinkpad_buffer_size = 0;
   }
 
-  pool = gst_vaapi_video_buffer_pool_new (plugin->display);
+  if (!gst_video_info_from_caps (&vi, caps))
+    goto error_invalid_caps;
+  gst_video_info_force_nv12_if_encoded (&vi);
+
+  if (!ensure_sinkpad_allocator (plugin, &vi))
+    goto error_create_allocator;
+
+  size = GST_VIDEO_INFO_SIZE (&vi);
+  gst_allocator_get_vaapi_image_size (plugin->sinkpad_allocator, &size);
+  pool = gst_vaapi_plugin_base_create_pool (plugin, caps, size, 0, 0,
+      GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META, plugin->sinkpad_allocator);
   if (!pool)
     goto error_create_pool;
-
-  gst_video_info_init (&vi);
-  gst_video_info_from_caps (&vi, caps);
-  if (GST_VIDEO_INFO_FORMAT (&vi) == GST_VIDEO_FORMAT_ENCODED) {
-    GST_DEBUG ("assume video buffer pool format is NV12");
-    gst_video_info_set_format (&vi, GST_VIDEO_FORMAT_NV12,
-        GST_VIDEO_INFO_WIDTH (&vi), GST_VIDEO_INFO_HEIGHT (&vi));
-  }
-  plugin->sinkpad_buffer_size = vi.size;
-
-  config = gst_buffer_pool_get_config (pool);
-  gst_buffer_pool_config_set_params (config, caps,
-      plugin->sinkpad_buffer_size, 0, 0);
-  gst_buffer_pool_config_add_option (config,
-      GST_BUFFER_POOL_OPTION_VAAPI_VIDEO_META);
-  gst_buffer_pool_config_add_option (config, GST_BUFFER_POOL_OPTION_VIDEO_META);
-  if (!gst_buffer_pool_set_config (pool, config))
-    goto error_pool_config;
   plugin->sinkpad_buffer_pool = pool;
+  plugin->sinkpad_buffer_size = size;
   return TRUE;
 
   /* ERRORS */
-error_create_pool:
+error_invalid_caps:
   {
-    GST_ERROR ("failed to create buffer pool");
+    GST_ERROR_OBJECT (plugin, "invalid caps %" GST_PTR_FORMAT, caps);
     return FALSE;
   }
-error_pool_config:
+error_create_allocator:
   {
-    GST_ERROR ("failed to reset buffer pool config");
-    gst_object_unref (pool);
+    GST_ERROR_OBJECT (plugin, "failed to create allocator");
+    return FALSE;
+  }
+error_create_pool:
+  {
+    /* error message already sent */
     return FALSE;
   }
 }
@@ -496,18 +704,19 @@ gst_vaapi_plugin_base_set_caps (GstVaapiPluginBase * plugin, GstCaps * incaps,
     GstCaps * outcaps)
 {
   if (incaps && incaps != plugin->sinkpad_caps) {
+    g_clear_object (&plugin->sinkpad_allocator);
     gst_caps_replace (&plugin->sinkpad_caps, incaps);
     if (!gst_video_info_from_caps (&plugin->sinkpad_info, incaps))
       return FALSE;
-    plugin->sinkpad_caps_changed = TRUE;
     plugin->sinkpad_caps_is_raw = !gst_caps_has_vaapi_surface (incaps);
   }
 
   if (outcaps && outcaps != plugin->srcpad_caps) {
+    g_clear_object (&plugin->srcpad_allocator);
     gst_caps_replace (&plugin->srcpad_caps, outcaps);
+    plugin_reset_texture_map (plugin);
     if (!gst_video_info_from_caps (&plugin->srcpad_info, outcaps))
       return FALSE;
-    plugin->srcpad_caps_changed = TRUE;
   }
 
   if (!ensure_sinkpad_buffer_pool (plugin, plugin->sinkpad_caps))
@@ -532,22 +741,17 @@ gst_vaapi_plugin_base_propose_allocation (GstVaapiPluginBase * plugin,
   gboolean need_pool;
 
   gst_query_parse_allocation (query, &caps, &need_pool);
+  if (!caps)
+    goto error_no_caps;
 
   if (need_pool) {
-    if (!caps)
-      goto error_no_caps;
     if (!ensure_sinkpad_buffer_pool (plugin, caps))
       return FALSE;
     gst_query_add_allocation_pool (query, plugin->sinkpad_buffer_pool,
         plugin->sinkpad_buffer_size, 0, 0);
 
     if (has_dmabuf_capable_peer (plugin, plugin->sinkpad)) {
-      GstStructure *const config =
-          gst_buffer_pool_get_config (plugin->sinkpad_buffer_pool);
-
-      gst_buffer_pool_config_add_option (config,
-          GST_BUFFER_POOL_OPTION_DMABUF_MEMORY);
-      if (!gst_buffer_pool_set_config (plugin->sinkpad_buffer_pool, config))
+      if (!set_dmabuf_allocator (plugin, plugin->sinkpad_buffer_pool, caps))
         goto error_pool_config;
     }
   }
@@ -569,17 +773,6 @@ error_pool_config:
   }
 }
 
-static inline gboolean
-gst_vaapi_plugin_base_set_pool_config (GstBufferPool * pool,
-    const gchar * option)
-{
-  GstStructure *config;
-
-  config = gst_buffer_pool_get_config (pool);
-  gst_buffer_pool_config_add_option (config, option);
-  return gst_buffer_pool_set_config (pool, config);
-}
-
 /**
  * gst_vaapi_plugin_base_decide_allocation:
  * @plugin: a #GstVaapiPluginBase
@@ -593,47 +786,35 @@ gst_vaapi_plugin_base_set_pool_config (GstBufferPool * pool,
  */
 gboolean
 gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
-    GstQuery * query, guint feature)
+    GstQuery * query)
 {
   GstCaps *caps = NULL;
   GstBufferPool *pool;
-  GstStructure *config;
   GstVideoInfo vi;
-  guint size, min, max;
+  guint size, min, max, pool_options;
   gboolean update_pool = FALSE;
-  gboolean has_video_meta = FALSE;
-  gboolean has_video_alignment = FALSE;
 #if (USE_GLX || USE_EGL)
-  gboolean has_texture_upload_meta = FALSE;
   guint idx;
 #endif
 
-  g_return_val_if_fail (plugin->display != NULL, FALSE);
-
   gst_query_parse_allocation (query, &caps, NULL);
-
-  /* We don't need any GL context beyond this point if not requested
-     so explicitly through GstVideoGLTextureUploadMeta */
-  gst_object_replace (&plugin->gl_context, NULL);
-
   if (!caps)
     goto error_no_caps;
 
-  if (!feature)
-    feature =
-        gst_vaapi_find_preferred_caps_feature (plugin->srcpad,
-        GST_VIDEO_FORMAT_ENCODED, NULL);
-
-  has_video_meta = gst_query_find_allocation_meta (query,
-      GST_VIDEO_META_API_TYPE, NULL);
+  pool_options = 0;
+  if (gst_query_find_allocation_meta (query, GST_VIDEO_META_API_TYPE, NULL))
+    pool_options |= GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_META;
 
 #if (USE_GLX || USE_EGL)
-  has_texture_upload_meta = gst_query_find_allocation_meta (query,
-      GST_VIDEO_GL_TEXTURE_UPLOAD_META_API_TYPE, &idx) &&
-      (feature == GST_VAAPI_CAPS_FEATURE_GL_TEXTURE_UPLOAD_META);
+  if (gst_query_find_allocation_meta (query,
+          GST_VIDEO_GL_TEXTURE_UPLOAD_META_API_TYPE, &idx) &&
+      gst_vaapi_caps_feature_contains (caps,
+          GST_VAAPI_CAPS_FEATURE_GL_TEXTURE_UPLOAD_META))
+    pool_options |= GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_GL_TEXTURE_UPLOAD;
 
 #if USE_GST_GL_HELPERS
-  if (has_texture_upload_meta) {
+  if (!plugin->gl_context &&
+      (pool_options & GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_GL_TEXTURE_UPLOAD)) {
     const GstStructure *params;
     GstObject *gl_context;
 
@@ -655,77 +836,46 @@ gst_vaapi_plugin_base_decide_allocation (GstVaapiPluginBase * plugin,
   if (!gst_vaapi_plugin_base_ensure_display (plugin))
     goto error_ensure_display;
 
-  gst_video_info_init (&vi);
-  gst_video_info_from_caps (&vi, caps);
-  if (GST_VIDEO_INFO_FORMAT (&vi) == GST_VIDEO_FORMAT_ENCODED)
-    gst_video_info_set_format (&vi, GST_VIDEO_FORMAT_I420,
-        GST_VIDEO_INFO_WIDTH (&vi), GST_VIDEO_INFO_HEIGHT (&vi));
+  if (!gst_video_info_from_caps (&vi, caps))
+    goto error_invalid_caps;
+  gst_video_info_force_nv12_if_encoded (&vi);
 
   if (gst_query_get_n_allocation_pools (query) > 0) {
     gst_query_parse_nth_allocation_pool (query, 0, &pool, &size, &min, &max);
     update_pool = TRUE;
-    size = MAX (size, vi.size);
+    size = MAX (size, GST_VIDEO_INFO_SIZE (&vi));
     if (pool) {
       /* Check whether downstream element proposed a bufferpool but did
          not provide a correct propose_allocation() implementation */
-      has_video_alignment = gst_buffer_pool_has_option (pool,
-          GST_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT);
+      if (gst_buffer_pool_has_option (pool,
+              GST_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT))
+        pool_options |= GST_VAAPI_VIDEO_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT;
+
+      /* GstVaapiVideoMeta is mandatory, and this implies VA surface memory */
+      if (!gst_buffer_pool_has_option (pool,
+              GST_BUFFER_POOL_OPTION_VAAPI_VIDEO_META)) {
+        GST_INFO_OBJECT (plugin, "ignoring non-VAAPI pool: %" GST_PTR_FORMAT,
+            pool);
+        g_clear_object (&pool);
+      }
     }
   } else {
     pool = NULL;
-    size = vi.size;
+    size = GST_VIDEO_INFO_SIZE (&vi);
     min = max = 0;
   }
 
-  /* GstVaapiVideoMeta is mandatory, and this implies VA surface memory */
-  if (!pool || !gst_buffer_pool_has_option (pool,
-          GST_BUFFER_POOL_OPTION_VAAPI_VIDEO_META)) {
-    GST_INFO_OBJECT (plugin, "%s. Making a new pool", pool == NULL ? "No pool" :
-        "Pool hasn't GstVaapiVideoMeta");
-    if (pool)
-      gst_object_unref (pool);
-    pool = gst_vaapi_video_buffer_pool_new (plugin->display);
+  if (!pool) {
+    if (!ensure_srcpad_allocator (plugin, &vi))
+      goto error_create_allocator;
+    /* Update video size with allocator's image size */
+    gst_allocator_get_vaapi_image_size (plugin->srcpad_allocator, &size);
+    pool = gst_vaapi_plugin_base_create_pool (plugin, caps, size, min, max,
+        pool_options, plugin->srcpad_allocator);
     if (!pool)
       goto error_create_pool;
-
-    config = gst_buffer_pool_get_config (pool);
-    gst_buffer_pool_config_set_params (config, caps, size, min, max);
-    gst_buffer_pool_config_add_option (config,
-        GST_BUFFER_POOL_OPTION_VAAPI_VIDEO_META);
-    if (!gst_buffer_pool_set_config (pool, config)) {
-      config = gst_buffer_pool_get_config (pool);
-
-      if (!gst_buffer_pool_config_validate_params (config, caps, size, min,
-              max)) {
-        gst_structure_free (config);
-        goto config_failed;
-      }
-
-      if (!gst_buffer_pool_set_config (pool, config))
-        goto config_failed;
-    }
   }
 
-  /* Check whether GstVideoMeta, or GstVideoAlignment, is needed (raw video) */
-  if (has_video_meta) {
-    if (!gst_vaapi_plugin_base_set_pool_config (pool,
-            GST_BUFFER_POOL_OPTION_VIDEO_META))
-      goto config_failed;
-  } else if (has_video_alignment) {
-    if (!gst_vaapi_plugin_base_set_pool_config (pool,
-            GST_BUFFER_POOL_OPTION_VIDEO_ALIGNMENT))
-      goto config_failed;
-  }
-
-  /* GstVideoGLTextureUploadMeta (OpenGL) */
-#if (USE_GLX || USE_EGL)
-  if (has_texture_upload_meta) {
-    if (!gst_vaapi_plugin_base_set_pool_config (pool,
-            GST_BUFFER_POOL_OPTION_VIDEO_GL_TEXTURE_UPLOAD_META))
-      goto config_failed;
-  }
-#endif
-
   if (update_pool)
     gst_query_set_nth_allocation_pool (query, 0, pool, size, min, max);
   else
@@ -741,24 +891,25 @@ error_no_caps:
     GST_ERROR_OBJECT (plugin, "no caps specified");
     return FALSE;
   }
+error_invalid_caps:
+  {
+    GST_ERROR_OBJECT (plugin, "invalid caps %" GST_PTR_FORMAT, caps);
+    return FALSE;
+  }
 error_ensure_display:
   {
     GST_ERROR_OBJECT (plugin, "failed to ensure display of type %d",
         plugin->display_type_req);
     return FALSE;
   }
-error_create_pool:
+error_create_allocator:
   {
-    GST_ERROR_OBJECT (plugin, "failed to create buffer pool");
+    GST_ERROR_OBJECT (plugin, "failed to create allocator");
     return FALSE;
   }
-config_failed:
+error_create_pool:
   {
-    if (pool)
-      gst_object_unref (pool);
-    GST_ELEMENT_ERROR (plugin, RESOURCE, SETTINGS,
-        ("Failed to configure the buffer pool"),
-        ("Configuration is most likely invalid, please report this issue."));
+    /* error message already sent */
     return FALSE;
   }
 }
@@ -766,7 +917,7 @@ config_failed:
 /**
  * gst_vaapi_plugin_base_get_input_buffer:
  * @plugin: a #GstVaapiPluginBase
- * @incaps: the sink pad (input) buffer
+ * @inbuf: the sink pad (input) buffer
  * @outbuf_ptr: the pointer to location to the VA surface backed buffer
  *
  * Acquires the sink pad (input) buffer as a VA surface backed
@@ -882,7 +1033,7 @@ error_bind_dma_buffer:
   }
 error_copy_buffer:
   {
-    GST_WARNING ("failed to upload buffer to VA surface");
+    GST_WARNING_OBJECT (plugin, "failed to upload buffer to VA surface");
     gst_buffer_unref (outbuf);
     return GST_FLOW_NOT_SUPPORTED;
   }
@@ -906,6 +1057,9 @@ gst_vaapi_plugin_base_set_gl_context (GstVaapiPluginBase * plugin,
   GstGLContext *const gl_context = GST_GL_CONTEXT (object);
   GstVaapiDisplayType display_type;
 
+  if (plugin->gl_context == object)
+    return;
+
   gst_object_replace (&plugin->gl_context, object);
 
   switch (gst_gl_context_get_gl_platform (gl_context)) {
@@ -923,6 +1077,7 @@ gst_vaapi_plugin_base_set_gl_context (GstVaapiPluginBase * plugin,
       display_type = plugin->display_type;
       break;
   }
+  GST_INFO_OBJECT (plugin, "GL context: %" GST_PTR_FORMAT, plugin->gl_context);
   gst_vaapi_plugin_base_set_display_type (plugin, display_type);
 #endif
 }
