@@ -39,6 +39,7 @@
 #include <gst/video/video.h>
 
 #include "gstvaapipostproc.h"
+#include "gstvaapipostprocutil.h"
 #include "gstvaapipluginutil.h"
 #include "gstvaapivideobuffer.h"
 #include "gstvaapivideobufferpool.h"
@@ -55,7 +56,6 @@ GST_DEBUG_CATEGORY_STATIC (gst_debug_vaapipostproc);
 static const char gst_vaapipostproc_sink_caps_str[] =
   GST_VAAPI_MAKE_SURFACE_CAPS ", "
   GST_CAPS_INTERLACED_MODES "; "
-  GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS "; "
   GST_VIDEO_CAPS_MAKE (GST_VIDEO_FORMATS_ALL) ", "
    GST_CAPS_INTERLACED_MODES;
 /* *INDENT-ON* */
@@ -64,7 +64,9 @@ static const char gst_vaapipostproc_sink_caps_str[] =
 static const char gst_vaapipostproc_src_caps_str[] =
   GST_VAAPI_MAKE_SURFACE_CAPS ", "
   GST_CAPS_INTERLACED_FALSE "; "
+#if (USE_GLX || USE_EGL)
   GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS "; "
+#endif
   GST_VIDEO_CAPS_MAKE (GST_VIDEO_FORMATS_ALL) ", "
   GST_CAPS_INTERLACED_FALSE;
 /* *INDENT-ON* */
@@ -117,10 +119,6 @@ enum
   PROP_SKIN_TONE_ENHANCEMENT,
 };
 
-#define DEFAULT_FORMAT                  GST_VIDEO_FORMAT_ENCODED
-#define DEFAULT_DEINTERLACE_MODE        GST_VAAPI_DEINTERLACE_MODE_AUTO
-#define DEFAULT_DEINTERLACE_METHOD      GST_VAAPI_DEINTERLACE_METHOD_BOB
-
 #define GST_VAAPI_TYPE_DEINTERLACE_MODE \
     gst_vaapi_deinterlace_mode_get_type()
 
@@ -346,7 +344,8 @@ should_deinterlace_buffer (GstVaapiPostproc * postproc, GstBuffer * buf)
         return TRUE;
       break;
     default:
-      GST_ERROR ("unhandled \"interlace-mode\", disabling deinterlacing");
+      GST_ERROR_OBJECT (postproc,
+          "unhandled \"interlace-mode\", disabling deinterlacing");
       break;
   }
   return FALSE;
@@ -375,12 +374,12 @@ create_output_buffer (GstVaapiPostproc * postproc)
   /* ERRORS */
 error_activate_pool:
   {
-    GST_ERROR ("failed to activate output video buffer pool");
+    GST_ERROR_OBJECT (postproc, "failed to activate output video buffer pool");
     return NULL;
   }
 error_create_buffer:
   {
-    GST_ERROR ("failed to create output video buffer");
+    GST_ERROR_OBJECT (postproc, "failed to create output video buffer");
     return NULL;
   }
 }
@@ -471,6 +470,128 @@ set_best_deint_method (GstVaapiPostproc * postproc, guint flags,
   return success;
 }
 
+static gboolean
+check_filter_update (GstVaapiPostproc * postproc)
+{
+  guint filter_flag = postproc->flags;
+  guint op_flag;
+  gint i;
+
+  if (!postproc->has_vpp)
+    return FALSE;
+
+  for (i = GST_VAAPI_FILTER_OP_DENOISE; i <= GST_VAAPI_FILTER_OP_SKINTONE; i++) {
+    op_flag = (filter_flag >> i) & 1;
+    if (op_flag)
+      return TRUE;
+  }
+
+  return FALSE;
+}
+
+static gboolean
+update_filter (GstVaapiPostproc * postproc)
+{
+  /* Validate filters */
+  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_FORMAT) &&
+      !gst_vaapi_filter_set_format (postproc->filter, postproc->format))
+    return FALSE;
+
+  if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_DENOISE) {
+    if (!gst_vaapi_filter_set_denoising_level (postproc->filter,
+            postproc->denoise_level))
+      return FALSE;
+
+    if (gst_vaapi_filter_get_denoising_level_default (postproc->filter) ==
+        postproc->denoise_level)
+      postproc->flags &= ~(GST_VAAPI_POSTPROC_FLAG_DENOISE);
+  }
+
+  if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_SHARPEN) {
+    if (!gst_vaapi_filter_set_sharpening_level (postproc->filter,
+            postproc->sharpen_level))
+      return FALSE;
+
+    if (gst_vaapi_filter_get_sharpening_level_default (postproc->filter) ==
+        postproc->sharpen_level)
+      postproc->flags &= ~(GST_VAAPI_POSTPROC_FLAG_SHARPEN);
+  }
+
+  if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_HUE) {
+    if (!gst_vaapi_filter_set_hue (postproc->filter, postproc->hue))
+      return FALSE;
+
+    if (gst_vaapi_filter_get_hue_default (postproc->filter) == postproc->hue)
+      postproc->flags &= ~(GST_VAAPI_POSTPROC_FLAG_HUE);
+  }
+
+  if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_SATURATION) {
+    if (!gst_vaapi_filter_set_saturation (postproc->filter,
+            postproc->saturation))
+      return FALSE;
+
+    if (gst_vaapi_filter_get_saturation_default (postproc->filter) ==
+        postproc->saturation)
+      postproc->flags &= ~(GST_VAAPI_POSTPROC_FLAG_SATURATION);
+  }
+
+  if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_BRIGHTNESS) {
+    if (!gst_vaapi_filter_set_brightness (postproc->filter,
+            postproc->brightness))
+      return FALSE;
+
+    if (gst_vaapi_filter_get_brightness_default (postproc->filter) ==
+        postproc->brightness)
+      postproc->flags &= ~(GST_VAAPI_POSTPROC_FLAG_BRIGHTNESS);
+  }
+
+  if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_CONTRAST) {
+    if (!gst_vaapi_filter_set_contrast (postproc->filter, postproc->contrast))
+      return FALSE;
+
+    if (gst_vaapi_filter_get_contrast_default (postproc->filter) ==
+        postproc->contrast)
+      postproc->flags &= ~(GST_VAAPI_POSTPROC_FLAG_CONTRAST);
+  }
+
+  if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_SCALE) {
+    if (!gst_vaapi_filter_set_scaling (postproc->filter,
+            postproc->scale_method))
+      return FALSE;
+
+    if (gst_vaapi_filter_get_scaling_default (postproc->filter) ==
+        postproc->scale_method)
+      postproc->flags &= ~(GST_VAAPI_POSTPROC_FLAG_SCALE);
+  }
+
+  if (postproc->flags & GST_VAAPI_POSTPROC_FLAG_SKINTONE) {
+    if (!gst_vaapi_filter_set_skintone (postproc->filter,
+            postproc->skintone_enhance))
+      return FALSE;
+
+    if (gst_vaapi_filter_get_skintone_default (postproc->filter) ==
+        postproc->skintone_enhance)
+      postproc->flags &= ~(GST_VAAPI_POSTPROC_FLAG_SKINTONE);
+  }
+
+  return TRUE;
+}
+
+static void
+gst_vaapipostproc_set_passthrough (GstBaseTransform * trans)
+{
+  GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
+  gboolean filter_updated = FALSE;
+
+  if (check_filter_update (postproc) && update_filter (postproc)) {
+    /* check again if changed value is default */
+    filter_updated = check_filter_update (postproc);
+  }
+
+  gst_base_transform_set_passthrough (trans, postproc->same_caps
+      && !filter_updated);
+}
+
 static GstFlowReturn
 gst_vaapipostproc_process_vpp (GstBaseTransform * trans, GstBuffer * inbuf,
     GstBuffer * outbuf)
@@ -491,46 +612,6 @@ gst_vaapipostproc_process_vpp (GstBaseTransform * trans, GstBuffer * inbuf,
   GstVaapiRectangle *crop_rect = NULL;
   GstVaapiRectangle tmp_rect;
 
-  /* Validate filters */
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_FORMAT) &&
-      !gst_vaapi_filter_set_format (postproc->filter, postproc->format))
-    return GST_FLOW_NOT_SUPPORTED;
-
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_DENOISE) &&
-      !gst_vaapi_filter_set_denoising_level (postproc->filter,
-          postproc->denoise_level))
-    return GST_FLOW_NOT_SUPPORTED;
-
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_SHARPEN) &&
-      !gst_vaapi_filter_set_sharpening_level (postproc->filter,
-          postproc->sharpen_level))
-    return GST_FLOW_NOT_SUPPORTED;
-
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_HUE) &&
-      !gst_vaapi_filter_set_hue (postproc->filter, postproc->hue))
-    return GST_FLOW_NOT_SUPPORTED;
-
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_SATURATION) &&
-      !gst_vaapi_filter_set_saturation (postproc->filter, postproc->saturation))
-    return GST_FLOW_NOT_SUPPORTED;
-
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_BRIGHTNESS) &&
-      !gst_vaapi_filter_set_brightness (postproc->filter, postproc->brightness))
-    return GST_FLOW_NOT_SUPPORTED;
-
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_CONTRAST) &&
-      !gst_vaapi_filter_set_contrast (postproc->filter, postproc->contrast))
-    return GST_FLOW_NOT_SUPPORTED;
-
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_SCALE) &&
-      !gst_vaapi_filter_set_scaling (postproc->filter, postproc->scale_method))
-    return GST_FLOW_NOT_SUPPORTED;
-
-  if ((postproc->flags & GST_VAAPI_POSTPROC_FLAG_SKINTONE) &&
-      !gst_vaapi_filter_set_skintone (postproc->filter,
-          postproc->skintone_enhance))
-    return GST_FLOW_NOT_SUPPORTED;
-
   inbuf_meta = gst_buffer_get_vaapi_video_meta (inbuf);
   if (!inbuf_meta)
     goto error_invalid_buffer;
@@ -643,13 +724,15 @@ gst_vaapipostproc_process_vpp (GstBaseTransform * trans, GstBuffer * inbuf,
   if (!outbuf_meta)
     goto error_create_meta;
 
-  proxy =
-      gst_vaapi_surface_proxy_new_from_pool (GST_VAAPI_SURFACE_POOL
-      (postproc->filter_pool));
-  if (!proxy)
-    goto error_create_proxy;
-  gst_vaapi_video_meta_set_surface_proxy (outbuf_meta, proxy);
-  gst_vaapi_surface_proxy_unref (proxy);
+  if (!gst_vaapi_video_meta_get_surface_proxy (outbuf_meta)) {
+    proxy =
+        gst_vaapi_surface_proxy_new_from_pool (GST_VAAPI_SURFACE_POOL
+        (postproc->filter_pool));
+    if (!proxy)
+      goto error_create_proxy;
+    gst_vaapi_video_meta_set_surface_proxy (outbuf_meta, proxy);
+    gst_vaapi_surface_proxy_unref (proxy);
+  }
 
   if (deint) {
     deint_flags = (tff ? 0 : GST_VAAPI_DEINTERLACE_FLAG_TOPFIELD);
@@ -690,42 +773,43 @@ gst_vaapipostproc_process_vpp (GstBaseTransform * trans, GstBuffer * inbuf,
   /* ERRORS */
 error_invalid_buffer:
   {
-    GST_ERROR ("failed to validate source buffer");
+    GST_ERROR_OBJECT (postproc, "failed to validate source buffer");
     return GST_FLOW_ERROR;
   }
 error_create_buffer:
   {
-    GST_ERROR ("failed to create output buffer");
+    GST_ERROR_OBJECT (postproc, "failed to create output buffer");
     return GST_FLOW_ERROR;
   }
 error_create_meta:
   {
-    GST_ERROR ("failed to create new output buffer meta");
+    GST_ERROR_OBJECT (postproc, "failed to create new output buffer meta");
     gst_buffer_replace (&fieldbuf, NULL);
     return GST_FLOW_ERROR;
   }
 error_create_proxy:
   {
-    GST_ERROR ("failed to create surface proxy from pool");
+    GST_ERROR_OBJECT (postproc, "failed to create surface proxy from pool");
     gst_buffer_replace (&fieldbuf, NULL);
     return GST_FLOW_ERROR;
   }
 error_op_deinterlace:
   {
-    GST_ERROR ("failed to apply deinterlacing filter");
+    GST_ERROR_OBJECT (postproc, "failed to apply deinterlacing filter");
     gst_buffer_replace (&fieldbuf, NULL);
     return GST_FLOW_NOT_SUPPORTED;
   }
 error_process_vpp:
   {
-    GST_ERROR ("failed to apply VPP filters (error %d)", status);
+    GST_ERROR_OBJECT (postproc, "failed to apply VPP filters (error %d)",
+        status);
     gst_buffer_replace (&fieldbuf, NULL);
     return GST_FLOW_ERROR;
   }
 error_push_buffer:
   {
     if (ret != GST_FLOW_FLUSHING)
-      GST_ERROR ("failed to push output buffer to video sink");
+      GST_ERROR_OBJECT (postproc, "failed to push output buffer to video sink");
     return GST_FLOW_ERROR;
   }
 }
@@ -791,18 +875,18 @@ gst_vaapipostproc_process (GstBaseTransform * trans, GstBuffer * inbuf,
   /* ERRORS */
 error_invalid_buffer:
   {
-    GST_ERROR ("failed to validate source buffer");
+    GST_ERROR_OBJECT (postproc, "failed to validate source buffer");
     return GST_FLOW_ERROR;
   }
 error_create_buffer:
   {
-    GST_ERROR ("failed to create output buffer");
+    GST_ERROR_OBJECT (postproc, "failed to create output buffer");
     return GST_FLOW_EOS;
   }
 error_push_buffer:
   {
     if (ret != GST_FLOW_FLUSHING)
-      GST_ERROR ("failed to push output buffer to video sink");
+      GST_ERROR_OBJECT (postproc, "failed to push output buffer to video sink");
     return GST_FLOW_EOS;
   }
 }
@@ -811,6 +895,7 @@ static GstFlowReturn
 gst_vaapipostproc_passthrough (GstBaseTransform * trans, GstBuffer * inbuf,
     GstBuffer * outbuf)
 {
+  GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
   GstVaapiVideoMeta *meta;
 
   /* No video processing needed, simply copy buffer metadata */
@@ -818,49 +903,26 @@ gst_vaapipostproc_passthrough (GstBaseTransform * trans, GstBuffer * inbuf,
   if (!meta)
     goto error_invalid_buffer;
 
-  append_output_buffer_metadata (GST_VAAPIPOSTPROC (trans), outbuf, inbuf,
+  append_output_buffer_metadata (postproc, outbuf, inbuf,
       GST_BUFFER_COPY_TIMESTAMPS);
   return GST_FLOW_OK;
 
   /* ERRORS */
 error_invalid_buffer:
   {
-    GST_ERROR ("failed to validate source buffer");
+    GST_ERROR_OBJECT (postproc, "failed to validate source buffer");
     return GST_FLOW_ERROR;
   }
 }
 
-static gboolean
-is_deinterlace_enabled (GstVaapiPostproc * postproc, GstVideoInfo * vip)
-{
-  gboolean deinterlace;
-
-  switch (postproc->deinterlace_mode) {
-    case GST_VAAPI_DEINTERLACE_MODE_AUTO:
-      deinterlace = GST_VIDEO_INFO_IS_INTERLACED (vip);
-      break;
-    case GST_VAAPI_DEINTERLACE_MODE_INTERLACED:
-      deinterlace = TRUE;
-      break;
-    default:
-      deinterlace = FALSE;
-      break;
-  }
-  return deinterlace;
-}
-
 static gboolean
 video_info_changed (GstVideoInfo * old_vip, GstVideoInfo * new_vip)
 {
-  if (GST_VIDEO_INFO_FORMAT (old_vip) != GST_VIDEO_INFO_FORMAT (new_vip))
+  if (gst_video_info_changed (old_vip, new_vip))
     return TRUE;
   if (GST_VIDEO_INFO_INTERLACE_MODE (old_vip) !=
       GST_VIDEO_INFO_INTERLACE_MODE (new_vip))
     return TRUE;
-  if (GST_VIDEO_INFO_WIDTH (old_vip) != GST_VIDEO_INFO_WIDTH (new_vip))
-    return TRUE;
-  if (GST_VIDEO_INFO_HEIGHT (old_vip) != GST_VIDEO_INFO_HEIGHT (new_vip))
-    return TRUE;
   return FALSE;
 }
 
@@ -934,26 +996,27 @@ ensure_allowed_sinkpad_caps (GstVaapiPostproc * postproc)
   if (postproc->allowed_sinkpad_caps)
     return TRUE;
 
+  if (!GST_VAAPI_PLUGIN_BASE_DISPLAY (postproc))
+    return FALSE;
+
   /* Create VA caps */
   out_caps = gst_caps_from_string (GST_VAAPI_MAKE_SURFACE_CAPS ", "
       GST_CAPS_INTERLACED_MODES);
   if (!out_caps) {
-    GST_ERROR_OBJECT (postproc, "failed to create VA sink caps");
+    GST_WARNING_OBJECT (postproc, "failed to create VA sink caps");
     return FALSE;
   }
 
-  /* Append raw video caps */
-  if (gst_vaapipostproc_ensure_display (postproc)) {
-    raw_caps =
-        gst_vaapi_plugin_base_get_allowed_raw_caps (GST_VAAPI_PLUGIN_BASE
-        (postproc));
-    if (raw_caps) {
-      out_caps = gst_caps_make_writable (out_caps);
-      gst_caps_append (out_caps, gst_caps_copy (raw_caps));
-    } else {
-      GST_WARNING_OBJECT (postproc, "failed to create YUV sink caps");
-    }
+  raw_caps = gst_vaapi_plugin_base_get_allowed_raw_caps
+      (GST_VAAPI_PLUGIN_BASE (postproc));
+  if (!raw_caps) {
+    gst_caps_unref (out_caps);
+    GST_WARNING_OBJECT (postproc, "failed to create YUV sink caps");
+    return FALSE;
   }
+
+  out_caps = gst_caps_make_writable (out_caps);
+  gst_caps_append (out_caps, gst_caps_copy (raw_caps));
   postproc->allowed_sinkpad_caps = out_caps;
 
   /* XXX: append VA/VPP filters */
@@ -1011,7 +1074,7 @@ ensure_allowed_srcpad_caps (GstVaapiPostproc * postproc)
   /* Create initial caps from pad template */
   out_caps = gst_caps_from_string (gst_vaapipostproc_src_caps_str);
   if (!out_caps) {
-    GST_ERROR ("failed to create VA src caps");
+    GST_ERROR_OBJECT (postproc, "failed to create VA src caps");
     return FALSE;
   }
 
@@ -1020,147 +1083,83 @@ ensure_allowed_srcpad_caps (GstVaapiPostproc * postproc)
   return postproc->allowed_srcpad_caps != NULL;
 }
 
-static void
-find_best_size (GstVaapiPostproc * postproc, GstVideoInfo * vip,
-    guint * width_ptr, guint * height_ptr)
-{
-  guint width, height;
-
-  width = GST_VIDEO_INFO_WIDTH (vip);
-  height = GST_VIDEO_INFO_HEIGHT (vip);
-  if (postproc->width && postproc->height) {
-    width = postproc->width;
-    height = postproc->height;
-  } else if (postproc->keep_aspect) {
-    const gdouble ratio = (gdouble) width / height;
-    if (postproc->width) {
-      width = postproc->width;
-      height = postproc->width / ratio;
-    } else if (postproc->height) {
-      height = postproc->height;
-      width = postproc->height * ratio;
-    }
-  } else if (postproc->width)
-    width = postproc->width;
-  else if (postproc->height)
-    height = postproc->height;
-
-  *width_ptr = width;
-  *height_ptr = height;
-}
-
 static GstCaps *
 gst_vaapipostproc_transform_caps_impl (GstBaseTransform * trans,
     GstPadDirection direction, GstCaps * caps)
 {
   GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
-  GstVideoInfo vi;
-  GstVideoFormat out_format;
-  GstCaps *out_caps;
-  GstVaapiCapsFeature feature;
-  const gchar *feature_str;
-  guint width, height;
 
   /* Generate the sink pad caps, that could be fixated afterwards */
   if (direction == GST_PAD_SRC) {
     if (!ensure_allowed_sinkpad_caps (postproc))
-      return NULL;
+      return gst_caps_from_string (gst_vaapipostproc_sink_caps_str);
     return gst_caps_ref (postproc->allowed_sinkpad_caps);
   }
 
-  /* Generate complete set of src pad caps if non-fixated sink pad
-     caps are provided */
-  if (!gst_caps_is_fixed (caps)) {
-    if (!ensure_allowed_srcpad_caps (postproc))
-      return NULL;
-    return gst_caps_ref (postproc->allowed_srcpad_caps);
-  }
-
-  /* Generate the expected src pad caps, from the current fixated
-     sink pad caps */
-  if (!gst_video_info_from_caps (&vi, caps))
+  /* Generate complete set of src pad caps */
+  if (!ensure_allowed_srcpad_caps (postproc))
     return NULL;
-
-  // Set double framerate in interlaced mode
-  if (is_deinterlace_enabled (postproc, &vi)) {
-    gint fps_n = GST_VIDEO_INFO_FPS_N (&vi);
-    gint fps_d = GST_VIDEO_INFO_FPS_D (&vi);
-    if (!gst_util_fraction_multiply (fps_n, fps_d, 2, 1, &fps_n, &fps_d))
-      return NULL;
-    GST_VIDEO_INFO_FPS_N (&vi) = fps_n;
-    GST_VIDEO_INFO_FPS_D (&vi) = fps_d;
-  }
-  // Signal the other pad that we only generate progressive frames
-  GST_VIDEO_INFO_INTERLACE_MODE (&vi) = GST_VIDEO_INTERLACE_MODE_PROGRESSIVE;
-
-  // Update size from user-specified parameters
-  find_best_size (postproc, &vi, &width, &height);
-
-  // Update format from user-specified parameters
-  /* XXX: this is a workaround until auto-plugging is fixed when
-   * format=ENCODED + memory:VASurface caps feature are provided.
-   * use the downstream negotiated video format as the output format
-   * if the user didn't explicitly ask for colorspace conversion.
-   * Use a filter caps which contain all raw video formats, (excluding
-   * GST_VIDEO_FORMAT_ENCODED) */
-  if (postproc->format != DEFAULT_FORMAT)
-    out_format = postproc->format;
-  else {
-    GstCaps *peer_caps;
-    GstVideoInfo peer_vi;
-    peer_caps =
-        gst_pad_peer_query_caps (GST_BASE_TRANSFORM_SRC_PAD (trans),
-        postproc->allowed_srcpad_caps);
-    if (gst_caps_is_any (peer_caps) || gst_caps_is_empty (peer_caps))
-      return peer_caps;
-    if (!gst_caps_is_fixed (peer_caps))
-      peer_caps = gst_caps_fixate (peer_caps);
-    gst_video_info_from_caps (&peer_vi, peer_caps);
-    out_format = GST_VIDEO_INFO_FORMAT (&peer_vi);
-    if (peer_caps)
-      gst_caps_unref (peer_caps);
-  }
-
-  feature =
-      gst_vaapi_find_preferred_caps_feature (GST_BASE_TRANSFORM_SRC_PAD (trans),
-      out_format, &out_format);
-  gst_video_info_change_format (&vi, out_format, width, height);
-
-  out_caps = gst_video_info_to_caps (&vi);
-  if (!out_caps)
-    return NULL;
-
-  if (feature) {
-    feature_str = gst_vaapi_caps_feature_to_string (feature);
-    if (feature_str)
-      gst_caps_set_features (out_caps, 0,
-          gst_caps_features_new (feature_str, NULL));
-  }
-
-  /* we don't need to do format conversion if GL_TEXTURE_UPLOAD_META
-   * is negotiated */
-  if (feature != GST_VAAPI_CAPS_FEATURE_GL_TEXTURE_UPLOAD_META &&
-      postproc->format != out_format) {
-    postproc->format = out_format;
-  }
-  return out_caps;
+  return gst_vaapipostproc_transform_srccaps (postproc);
 }
 
 static GstCaps *
 gst_vaapipostproc_transform_caps (GstBaseTransform * trans,
     GstPadDirection direction, GstCaps * caps, GstCaps * filter)
 {
+  GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
   GstCaps *out_caps;
 
+  GST_DEBUG_OBJECT (trans,
+      "Transforming caps %" GST_PTR_FORMAT " in direction %s", caps,
+      (direction == GST_PAD_SINK) ? "sink" : "src");
+
+  g_mutex_lock (&postproc->postproc_lock);
   caps = gst_vaapipostproc_transform_caps_impl (trans, direction, caps);
+  g_mutex_unlock (&postproc->postproc_lock);
   if (caps && filter) {
     out_caps = gst_caps_intersect_full (caps, filter, GST_CAPS_INTERSECT_FIRST);
     gst_caps_unref (caps);
-    return out_caps;
+    caps = out_caps;
   }
+
+  GST_DEBUG_OBJECT (trans, "returning caps: %" GST_PTR_FORMAT, caps);
+
   return caps;
 }
 
+static GstCaps *
+gst_vaapipostproc_fixate_caps (GstBaseTransform * trans,
+    GstPadDirection direction, GstCaps * caps, GstCaps * othercaps)
+{
+  GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
+  GstCaps *outcaps = NULL;
+
+  GST_DEBUG_OBJECT (trans, "trying to fixate othercaps %" GST_PTR_FORMAT
+      " based on caps %" GST_PTR_FORMAT " in direction %s", othercaps, caps,
+      (direction == GST_PAD_SINK) ? "sink" : "src");
+
+  if (direction == GST_PAD_SRC) {
+    /* @TODO: we can do better */
+    othercaps = gst_caps_fixate (othercaps);
+    goto done;
+  }
+
+  g_mutex_lock (&postproc->postproc_lock);
+  if ((outcaps = gst_vaapipostproc_fixate_srccaps (postproc, caps, othercaps)))
+    gst_caps_replace (&othercaps, outcaps);
+  g_mutex_unlock (&postproc->postproc_lock);
+
+  /* set passthrough according to caps changes or filter changes */
+  gst_vaapipostproc_set_passthrough (trans);
+
+done:
+  GST_DEBUG_OBJECT (trans, "fixated othercaps to %" GST_PTR_FORMAT, othercaps);
+  if (outcaps)
+    gst_caps_unref (outcaps);
+
+  return othercaps;
+}
+
 static gboolean
 gst_vaapipostproc_transform_size (GstBaseTransform * trans,
     GstPadDirection direction, GstCaps * caps, gsize size,
@@ -1198,7 +1197,7 @@ gst_vaapipostproc_transform (GstBaseTransform * trans, GstBuffer * inbuf,
       ret = gst_vaapipostproc_process_vpp (trans, buf, outbuf);
       if (ret != GST_FLOW_NOT_SUPPORTED)
         goto done;
-      GST_WARNING ("unsupported VPP filters. Disabling");
+      GST_WARNING_OBJECT (postproc, "unsupported VPP filters. Disabling");
     }
 
     /* Only append picture structure meta data (top/bottom field) */
@@ -1223,6 +1222,11 @@ gst_vaapipostproc_prepare_output_buffer (GstBaseTransform * trans,
 {
   GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
 
+  if (gst_base_transform_is_passthrough (trans)) {
+    *outbuf_ptr = inbuf;
+    return GST_FLOW_OK;
+  }
+
   *outbuf_ptr = create_output_buffer (postproc);
   return *outbuf_ptr ? GST_FLOW_OK : GST_FLOW_ERROR;
 }
@@ -1233,8 +1237,8 @@ ensure_srcpad_buffer_pool (GstVaapiPostproc * postproc, GstCaps * caps)
   GstVideoInfo vi;
   GstVaapiVideoPool *pool;
 
-  gst_video_info_init (&vi);
-  gst_video_info_from_caps (&vi, caps);
+  if (!gst_video_info_from_caps (&vi, caps))
+    return FALSE;
   gst_video_info_change_format (&vi, postproc->format,
       GST_VIDEO_INFO_WIDTH (&vi), GST_VIDEO_INFO_HEIGHT (&vi));
 
@@ -1271,40 +1275,53 @@ gst_vaapipostproc_set_caps (GstBaseTransform * trans, GstCaps * caps,
   GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
   gboolean caps_changed = FALSE;
   GstVideoInfo vinfo;
+  gboolean ret = FALSE;
 
+  g_mutex_lock (&postproc->postproc_lock);
   if (!gst_vaapipostproc_update_sink_caps (postproc, caps, &caps_changed))
-    return FALSE;
+    goto done;
   /* HACK: This is a workaround to deal with the va-intel-driver for non-native
    * formats while doing advanced deinterlacing. The format of reference surfaces must
    * be same as the format used by the driver internally for motion adaptive
    * deinterlacing and motion compensated deinterlacing */
-  gst_video_info_from_caps (&vinfo, caps);
+  if (!gst_video_info_from_caps (&vinfo, caps))
+    goto done;
   if (deint_method_is_advanced (postproc->deinterlace_method)
       && !is_native_video_format (GST_VIDEO_INFO_FORMAT (&vinfo))) {
     GST_WARNING_OBJECT (postproc,
         "Advanced deinterlacing requires the native video formats used by the driver internally");
-    return FALSE;
+    goto done;
   }
   if (!gst_vaapipostproc_update_src_caps (postproc, out_caps, &caps_changed))
-    return FALSE;
+    goto done;
 
   if (caps_changed) {
     gst_vaapipostproc_destroy (postproc);
     if (!gst_vaapipostproc_create (postproc))
-      return FALSE;
+      goto done;
     if (!gst_vaapi_plugin_base_set_caps (GST_VAAPI_PLUGIN_BASE (trans),
             caps, out_caps))
-      return FALSE;
+      goto done;
   }
 
   if (!ensure_srcpad_buffer_pool (postproc, out_caps))
-    return FALSE;
-  return TRUE;
+    goto done;
+
+  postproc->same_caps = gst_caps_is_equal (caps, out_caps);
+
+  /* set passthrough according to caps changes or filter changes */
+  gst_vaapipostproc_set_passthrough (trans);
+
+  ret = TRUE;
+
+done:
+  g_mutex_unlock (&postproc->postproc_lock);
+  return ret;
 }
 
 static gboolean
-gst_vaapipostproc_query (GstBaseTransform * trans, GstPadDirection direction,
-    GstQuery * query)
+gst_vaapipostproc_query (GstBaseTransform * trans,
+    GstPadDirection direction, GstQuery * query)
 {
   GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
 
@@ -1328,7 +1345,33 @@ gst_vaapipostproc_propose_allocation (GstBaseTransform * trans,
 {
   GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (trans);
   GstVaapiPluginBase *const plugin = GST_VAAPI_PLUGIN_BASE (trans);
+  GstCaps *allocation_caps;
+  GstStructure *structure;
+  gint allocation_width, allocation_height;
+  gint negotiated_width, negotiated_height;
 
+  negotiated_width = GST_VIDEO_INFO_WIDTH (&postproc->sinkpad_info);
+  negotiated_height = GST_VIDEO_INFO_HEIGHT (&postproc->sinkpad_info);
+
+  if (negotiated_width == 0 || negotiated_height == 0)
+    goto bail;
+
+  allocation_caps = NULL;
+  gst_query_parse_allocation (query, &allocation_caps, NULL);
+  if (!allocation_caps)
+    goto bail;
+
+  structure = gst_caps_get_structure (allocation_caps, 0);
+  if (!gst_structure_get_int (structure, "width", &allocation_width))
+    goto bail;
+  if (!gst_structure_get_int (structure, "height", &allocation_height))
+    goto bail;
+
+  if (allocation_width != negotiated_width
+      || allocation_height != negotiated_height)
+    postproc->flags |= GST_VAAPI_POSTPROC_FLAG_SIZE;
+
+bail:
   /* Let vaapidecode allocate the video buffers */
   if (postproc->get_va_surfaces)
     return FALSE;
@@ -1341,7 +1384,7 @@ static gboolean
 gst_vaapipostproc_decide_allocation (GstBaseTransform * trans, GstQuery * query)
 {
   return gst_vaapi_plugin_base_decide_allocation (GST_VAAPI_PLUGIN_BASE (trans),
-      query, 0);
+      query);
 }
 
 static void
@@ -1351,6 +1394,7 @@ gst_vaapipostproc_finalize (GObject * object)
 
   gst_vaapipostproc_destroy (postproc);
 
+  g_mutex_clear (&postproc->postproc_lock);
   gst_vaapi_plugin_base_finalize (GST_VAAPI_PLUGIN_BASE (postproc));
   G_OBJECT_CLASS (gst_vaapipostproc_parent_class)->finalize (object);
 }
@@ -1361,6 +1405,7 @@ gst_vaapipostproc_set_property (GObject * object,
 {
   GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (object);
 
+  g_mutex_lock (&postproc->postproc_lock);
   switch (prop_id) {
     case PROP_FORMAT:
       postproc->format = g_value_get_enum (value);
@@ -1416,6 +1461,10 @@ gst_vaapipostproc_set_property (GObject * object,
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
+  g_mutex_unlock (&postproc->postproc_lock);
+
+  if (check_filter_update (postproc))
+    gst_base_transform_reconfigure_src (GST_BASE_TRANSFORM (postproc));
 }
 
 static void
@@ -1424,6 +1473,7 @@ gst_vaapipostproc_get_property (GObject * object,
 {
   GstVaapiPostproc *const postproc = GST_VAAPIPOSTPROC (object);
 
+  g_mutex_lock (&postproc->postproc_lock);
   switch (prop_id) {
     case PROP_FORMAT:
       g_value_set_enum (value, postproc->format);
@@ -1471,6 +1521,7 @@ gst_vaapipostproc_get_property (GObject * object,
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
+  g_mutex_unlock (&postproc->postproc_lock);
 }
 
 static void
@@ -1479,7 +1530,6 @@ gst_vaapipostproc_class_init (GstVaapiPostprocClass * klass)
   GObjectClass *const object_class = G_OBJECT_CLASS (klass);
   GstElementClass *const element_class = GST_ELEMENT_CLASS (klass);
   GstBaseTransformClass *const trans_class = GST_BASE_TRANSFORM_CLASS (klass);
-  GstPadTemplate *pad_template;
   GPtrArray *filter_ops;
   GstVaapiFilterOpInfo *filter_op;
 
@@ -1493,6 +1543,7 @@ gst_vaapipostproc_class_init (GstVaapiPostprocClass * klass)
   object_class->get_property = gst_vaapipostproc_get_property;
   trans_class->start = gst_vaapipostproc_start;
   trans_class->stop = gst_vaapipostproc_stop;
+  trans_class->fixate_caps = gst_vaapipostproc_fixate_caps;
   trans_class->transform_caps = gst_vaapipostproc_transform_caps;
   trans_class->transform_size = gst_vaapipostproc_transform_size;
   trans_class->transform = gst_vaapipostproc_transform;
@@ -1511,12 +1562,12 @@ gst_vaapipostproc_class_init (GstVaapiPostprocClass * klass)
       GST_PLUGIN_DESC, "Gwenole Beauchesne <gwenole.beauchesne@intel.com>");
 
   /* sink pad */
-  pad_template = gst_static_pad_template_get (&gst_vaapipostproc_sink_factory);
-  gst_element_class_add_pad_template (element_class, pad_template);
+  gst_element_class_add_static_pad_template (element_class,
+      &gst_vaapipostproc_sink_factory);
 
   /* src pad */
-  pad_template = gst_static_pad_template_get (&gst_vaapipostproc_src_factory);
-  gst_element_class_add_pad_template (element_class, pad_template);
+  gst_element_class_add_static_pad_template (element_class,
+      &gst_vaapipostproc_src_factory);
 
   /**
    * GstVaapiPostproc:deinterlace-mode:
@@ -1740,6 +1791,7 @@ gst_vaapipostproc_init (GstVaapiPostproc * postproc)
   gst_vaapi_plugin_base_init (GST_VAAPI_PLUGIN_BASE (postproc),
       GST_CAT_DEFAULT);
 
+  g_mutex_init (&postproc->postproc_lock);
   postproc->format = DEFAULT_FORMAT;
   postproc->deinterlace_mode = DEFAULT_DEINTERLACE_MODE;
   postproc->deinterlace_method = DEFAULT_DEINTERLACE_METHOD;
@@ -1772,10 +1824,11 @@ typedef struct
 } ColorBalanceChannel;
 
 ColorBalanceChannel cb_channels[] = {
-  {GST_VAAPI_FILTER_OP_HUE, "VA_FILTER_HUE"},
-  {GST_VAAPI_FILTER_OP_SATURATION, "VA_FILTER_SATURATION"},
-  {GST_VAAPI_FILTER_OP_BRIGHTNESS, "VA_FILTER_BRIGHTNESS"},
-  {GST_VAAPI_FILTER_OP_CONTRAST, "VA_FILTER_CONTRAST"},
+  {
+      GST_VAAPI_FILTER_OP_HUE, "VA_FILTER_HUE"}, {
+      GST_VAAPI_FILTER_OP_SATURATION, "VA_FILTER_SATURATION"}, {
+      GST_VAAPI_FILTER_OP_BRIGHTNESS, "VA_FILTER_BRIGHTNESS"}, {
+      GST_VAAPI_FILTER_OP_CONTRAST, "VA_FILTER_CONTRAST"},
 };
 
 static void
@@ -1825,8 +1878,8 @@ gst_vaapipostproc_colorbalance_list_channels (GstColorBalance * balance)
 }
 
 static gfloat *
-cb_get_value_ptr (GstVaapiPostproc * postproc, GstColorBalanceChannel * channel,
-    GstVaapiPostprocFlags * flags)
+cb_get_value_ptr (GstVaapiPostproc * postproc,
+    GstColorBalanceChannel * channel, GstVaapiPostprocFlags * flags)
 {
   guint i;
   gfloat *ret = NULL;
@@ -1860,6 +1913,8 @@ gst_vaapipostproc_colorbalance_set_value (GstColorBalance * balance,
     *var = new_val;
     postproc->flags |= flags;
     gst_color_balance_value_changed (balance, channel, value);
+    if (check_filter_update (postproc))
+      gst_base_transform_reconfigure_src (GST_BASE_TRANSFORM (postproc));
     return;
   }
 
