@@ -64,11 +64,11 @@ GST_DEBUG_CATEGORY_STATIC (gst_debug_vaapisink);
 /* Default template */
 /* *INDENT-OFF* */
 static const char gst_vaapisink_sink_caps_str[] =
-    GST_VAAPI_MAKE_ENC_SURFACE_CAPS ";"
+    GST_VAAPI_MAKE_SURFACE_CAPS ";"
     GST_VIDEO_CAPS_MAKE_WITH_FEATURES (
         GST_CAPS_FEATURE_MEMORY_VAAPI_SURFACE ","
             GST_CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION,
-        "{ ENCODED, NV12, I420, YV12 }") ";"
+        "{ ENCODED, NV12, I420, YV12, P010_10LE }") ";"
     GST_VIDEO_CAPS_MAKE_WITH_FEATURES (
         GST_CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION,
         GST_VIDEO_FORMATS_ALL) ";"
@@ -1114,6 +1114,26 @@ gst_vaapisink_ensure_colorbalance (GstVaapiSink * sink)
   return cb_sync_values_to_display (sink, GST_VAAPI_PLUGIN_BASE_DISPLAY (sink));
 }
 
+
+static void
+gst_vaapisink_set_rotation (GstVaapiSink * sink, GstVaapiRotation rotation,
+    gboolean from_tag)
+{
+  GST_OBJECT_LOCK (sink);
+
+  if (from_tag)
+    sink->rotation_tag = rotation;
+  else
+    sink->rotation_prop = rotation;
+
+  if (sink->rotation_prop == GST_VAAPI_ROTATION_AUTOMATIC)
+    sink->rotation_req = sink->rotation_tag;
+  else
+    sink->rotation_req = sink->rotation_prop;
+
+  GST_OBJECT_UNLOCK (sink);
+}
+
 static gboolean
 gst_vaapisink_ensure_rotation (GstVaapiSink * sink,
     gboolean recalc_display_rect)
@@ -1217,7 +1237,7 @@ gst_vaapisink_get_caps_impl (GstBaseSink * base_sink)
   GstVaapiSink *const sink = GST_VAAPISINK_CAST (base_sink);
   GstCaps *out_caps, *raw_caps, *feature_caps;
   static const char surface_caps_str[] =
-      GST_VAAPI_MAKE_ENC_SURFACE_CAPS ";"
+      GST_VAAPI_MAKE_SURFACE_CAPS ";"
       GST_VIDEO_CAPS_MAKE_WITH_FEATURES (GST_CAPS_FEATURE_MEMORY_VAAPI_SURFACE
       "," GST_CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION,
       "{ ENCODED, NV12, I420, YV12 }");
@@ -1305,7 +1325,7 @@ gst_vaapisink_set_caps (GstBaseSink * base_sink, GstCaps * caps)
   GST_DEBUG ("video pixel-aspect-ratio %d/%d",
       sink->video_par_n, sink->video_par_d);
 
-  update_colorimetry (sink, &vip->colorimetry);
+  update_colorimetry (sink, &GST_VIDEO_INFO_COLORIMETRY (vip));
   gst_caps_replace (&sink->caps, caps);
 
   gst_vaapisink_ensure_colorbalance (sink);
@@ -1389,7 +1409,7 @@ gst_vaapisink_show_frame_unlocked (GstVaapiSink * sink, GstBuffer * src_buffer)
   gst_vaapisink_ensure_colorbalance (sink);
   gst_vaapisink_ensure_rotation (sink, TRUE);
 
-  GST_DEBUG ("render surface %" GST_VAAPI_ID_FORMAT,
+  GST_TRACE_OBJECT (sink, "render surface %" GST_VAAPI_ID_FORMAT,
       GST_VAAPI_ID_ARGS (gst_vaapi_surface_get_id (surface)));
 
   if (!surface_rect)
@@ -1429,17 +1449,22 @@ done:
   gst_buffer_unref (buffer);
   return ret;
 
+  /* ERRORS */
 error:
-  GST_ELEMENT_ERROR (sink, RESOURCE, WRITE,
-      ("Internal error: could not render surface"), (NULL));
-  ret = GST_FLOW_ERROR;
-  goto done;
+  {
+    GST_ELEMENT_ERROR (sink, RESOURCE, WRITE,
+        ("Internal error: could not render surface"), (NULL));
+    ret = GST_FLOW_ERROR;
+    goto done;
+  }
 
 no_surface:
-  /* No surface or surface proxy. That's very bad! */
-  GST_WARNING_OBJECT (sink, "could not get surface");
-  ret = GST_FLOW_ERROR;
-  goto done;
+  {
+    /* No surface or surface proxy. That's very bad! */
+    GST_WARNING_OBJECT (sink, "could not get surface");
+    ret = GST_FLOW_ERROR;
+    goto done;
+  }
 }
 
 static GstFlowReturn
@@ -1533,7 +1558,7 @@ gst_vaapisink_set_property (GObject * object,
       sink->view_id = g_value_get_int (value);
       break;
     case PROP_ROTATION:
-      sink->rotation_req = g_value_get_enum (value);
+      gst_vaapisink_set_rotation (sink, g_value_get_enum (value), FALSE);
       break;
     case PROP_FORCE_ASPECT_RATIO:
       sink->keep_aspect = g_value_get_boolean (value);
@@ -1616,6 +1641,51 @@ gst_vaapisink_unlock_stop (GstBaseSink * base_sink)
   return TRUE;
 }
 
+static gboolean
+gst_vaapisink_event (GstBaseSink * base_sink, GstEvent * event)
+{
+  gboolean res = TRUE;
+  GstTagList *taglist;
+  gchar *orientation;
+
+  GstVaapiSink *const sink = GST_VAAPISINK_CAST (base_sink);
+
+  GST_DEBUG_OBJECT (sink, "handling event %s", GST_EVENT_TYPE_NAME (event));
+
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_TAG:
+      gst_event_parse_tag (event, &taglist);
+
+      if (gst_tag_list_get_string (taglist, GST_TAG_IMAGE_ORIENTATION,
+              &orientation)) {
+        if (!g_strcmp0 ("rotate-0", orientation)) {
+          gst_vaapisink_set_rotation (sink, GST_VAAPI_ROTATION_0, TRUE);
+        } else if (!g_strcmp0 ("rotate-90", orientation)) {
+          gst_vaapisink_set_rotation (sink, GST_VAAPI_ROTATION_90, TRUE);
+        } else if (!g_strcmp0 ("rotate-180", orientation)) {
+          gst_vaapisink_set_rotation (sink, GST_VAAPI_ROTATION_180, TRUE);
+        } else if (!g_strcmp0 ("rotate-270", orientation)) {
+          gst_vaapisink_set_rotation (sink, GST_VAAPI_ROTATION_270, TRUE);
+        }
+
+        /* Do not support for flip yet.
+         * It should be implemented in the near future.
+         * See https://bugs.freedesktop.org/show_bug.cgi?id=90654
+         */
+        g_free (orientation);
+      }
+      break;
+    default:
+      break;
+  }
+
+  res =
+      GST_BASE_SINK_CLASS (gst_vaapisink_parent_class)->event (base_sink,
+      event);
+
+  return res;
+}
+
 static void
 gst_vaapisink_set_bus (GstElement * element, GstBus * bus)
 {
@@ -1638,7 +1708,6 @@ gst_vaapisink_class_init (GstVaapiSinkClass * klass)
   GstVideoSinkClass *const videosink_class = GST_VIDEO_SINK_CLASS (klass);
   GstVaapiPluginBaseClass *const base_plugin_class =
       GST_VAAPI_PLUGIN_BASE_CLASS (klass);
-  GstPadTemplate *pad_template;
 
   GST_DEBUG_CATEGORY_INIT (gst_debug_vaapisink,
       GST_PLUGIN_NAME, 0, GST_PLUGIN_DESC);
@@ -1659,6 +1728,7 @@ gst_vaapisink_class_init (GstVaapiSinkClass * klass)
   basesink_class->propose_allocation = gst_vaapisink_propose_allocation;
   basesink_class->unlock = gst_vaapisink_unlock;
   basesink_class->unlock_stop = gst_vaapisink_unlock_stop;
+  basesink_class->event = gst_vaapisink_event;
 
   videosink_class->show_frame = GST_DEBUG_FUNCPTR (gst_vaapisink_show_frame);
 
@@ -1668,8 +1738,8 @@ gst_vaapisink_class_init (GstVaapiSinkClass * klass)
       "VA-API sink", "Sink/Video", GST_PLUGIN_DESC,
       "Gwenole Beauchesne <gwenole.beauchesne@intel.com>");
 
-  pad_template = gst_static_pad_template_get (&gst_vaapisink_sink_factory);
-  gst_element_class_add_pad_template (element_class, pad_template);
+  gst_element_class_add_static_pad_template (element_class,
+      &gst_vaapisink_sink_factory);
 
   /**
    * GstVaapiSink:display:
@@ -1829,6 +1899,7 @@ gst_vaapisink_init (GstVaapiSink * sink)
   sink->handle_events = TRUE;
   sink->rotation = DEFAULT_ROTATION;
   sink->rotation_req = DEFAULT_ROTATION;
+  sink->rotation_tag = DEFAULT_ROTATION;
   sink->keep_aspect = TRUE;
   sink->signal_handoffs = DEFAULT_SIGNAL_HANDOFFS;
   gst_video_info_init (&sink->video_info);
