@@ -26,6 +26,7 @@
 #include <gst/vaapi/gstvaapisurfacepool.h>
 #include <gst/vaapi/gstvaapiimagepool.h>
 #include "gstvaapivideomemory.h"
+#include "gstvaapipluginutil.h"
 
 GST_DEBUG_CATEGORY_STATIC (gst_debug_vaapivideomemory);
 #define GST_CAT_DEFAULT gst_debug_vaapivideomemory
@@ -351,7 +352,8 @@ gst_vaapi_video_memory_reset_image (GstVaapiVideoMemory * mem)
     mem->image = NULL;
   }
 
-  /* Don't synchronize to surface, this shall have happened during unmaps */
+  /* Don't synchronize to surface, this shall have happened during
+   * unmaps */
   GST_VAAPI_VIDEO_MEMORY_FLAG_UNSET (mem,
       GST_VAAPI_VIDEO_MEMORY_FLAG_IMAGE_IS_CURRENT);
 }
@@ -434,29 +436,45 @@ gst_vaapi_video_memory_map (GstVaapiVideoMemory * mem, gsize maxsize,
 
   /* ERRORS */
 error_unsupported_map:
-  GST_ERROR ("unsupported map flags (0x%x)", flags);
-  return NULL;
+  {
+    GST_ERROR ("unsupported map flags (0x%x)", flags);
+    return NULL;
+  }
 error_unsupported_map_type:
-  GST_ERROR ("unsupported map type (%d)", mem->map_type);
-  return NULL;
+  {
+    GST_ERROR ("unsupported map type (%d)", mem->map_type);
+    return NULL;
+  }
 error_no_surface_proxy:
-  GST_ERROR ("failed to extract GstVaapiSurfaceProxy from video meta");
-  return NULL;
+  {
+    GST_ERROR ("failed to extract GstVaapiSurfaceProxy from video meta");
+    return NULL;
+  }
 error_no_surface:
-  GST_ERROR ("failed to extract VA surface from video buffer");
-  return NULL;
+  {
+    GST_ERROR ("failed to extract VA surface from video buffer");
+    return NULL;
+  }
 error_no_current_surface:
-  GST_ERROR ("failed to make surface current");
-  return NULL;
+  {
+    GST_ERROR ("failed to make surface current");
+    return NULL;
+  }
 error_no_image:
-  GST_ERROR ("failed to extract VA image from video buffer");
-  return NULL;
+  {
+    GST_ERROR ("failed to extract VA image from video buffer");
+    return NULL;
+  }
 error_no_current_image:
-  GST_ERROR ("failed to make image current");
-  return NULL;
+  {
+    GST_ERROR ("failed to make image current");
+    return NULL;
+  }
 error_map_image:
-  GST_ERROR ("failed to map VA image");
-  return NULL;
+  {
+    GST_ERROR ("failed to map VA image");
+    return NULL;
+  }
 }
 
 static void
@@ -480,8 +498,10 @@ gst_vaapi_video_memory_unmap (GstVaapiVideoMemory * mem)
 
   /* ERRORS */
 error_incompatible_map:
-  GST_ERROR ("incompatible map type (%d)", mem->map_type);
-  return;
+  {
+    GST_ERROR ("incompatible map type (%d)", mem->map_type);
+    return;
+  }
 }
 
 static GstVaapiVideoMemory *
@@ -516,14 +536,20 @@ gst_vaapi_video_memory_copy (GstVaapiVideoMemory * mem,
 
   /* ERRORS */
 error_no_current_surface:
-  GST_ERROR ("failed to make surface current");
-  return NULL;
+  {
+    GST_ERROR ("failed to make surface current");
+    return NULL;
+  }
 error_unsupported:
-  GST_ERROR ("failed to copy partial memory (unsupported operation)");
-  return NULL;
+  {
+    GST_ERROR ("failed to copy partial memory (unsupported operation)");
+    return NULL;
+  }
 error_allocate_memory:
-  GST_ERROR ("failed to allocate GstVaapiVideoMemory copy");
-  return NULL;
+  {
+    GST_ERROR ("failed to allocate GstVaapiVideoMemory copy");
+    return NULL;
+  }
 }
 
 static GstVaapiVideoMemory *
@@ -556,16 +582,6 @@ gst_vaapi_video_memory_is_span (GstVaapiVideoMemory * mem1,
 G_DEFINE_TYPE (GstVaapiVideoAllocator,
     gst_vaapi_video_allocator, GST_TYPE_ALLOCATOR);
 
-static GstMemory *
-gst_vaapi_video_allocator_alloc (GstAllocator * allocator, gsize size,
-    GstAllocationParams * params)
-{
-  g_warning ("use gst_vaapi_video_memory_new() to allocate from "
-      "GstVaapiVideoMemory allocator");
-
-  return NULL;
-}
-
 static void
 gst_vaapi_video_allocator_free (GstAllocator * allocator, GstMemory * mem)
 {
@@ -594,7 +610,6 @@ gst_vaapi_video_allocator_class_init (GstVaapiVideoAllocatorClass * klass)
       "vaapivideomemory", 0, "VA-API video memory allocator");
 
   object_class->finalize = gst_vaapi_video_allocator_finalize;
-  allocator_class->alloc = gst_vaapi_video_allocator_alloc;
   allocator_class->free = gst_vaapi_video_allocator_free;
 }
 
@@ -614,6 +629,8 @@ gst_vaapi_video_allocator_init (GstVaapiVideoAllocator * allocator)
       gst_vaapi_video_memory_share;
   base_allocator->mem_is_span = (GstMemoryIsSpanFunction)
       gst_vaapi_video_memory_is_span;
+
+  GST_OBJECT_FLAG_SET (allocator, GST_ALLOCATOR_FLAG_CUSTOM_ALLOC);
 }
 
 static gboolean
@@ -714,12 +731,8 @@ allocator_configure_image_info (GstVaapiDisplay * display,
   }
 
   vinfo = &allocator->video_info;
-
-  if (GST_VIDEO_INFO_FORMAT (vinfo) == GST_VIDEO_FORMAT_ENCODED)
-    gst_video_info_set_format (&allocator->image_info, GST_VIDEO_FORMAT_I420,
-        GST_VIDEO_INFO_WIDTH (vinfo), GST_VIDEO_INFO_HEIGHT (vinfo));
-  else
-    allocator->image_info = *vinfo;
+  allocator->image_info = *vinfo;
+  gst_video_info_force_nv12_if_encoded (&allocator->image_info);
 
   image = new_image (display, &allocator->image_info);
   if (!image)
@@ -737,7 +750,7 @@ bail:
 
 GstAllocator *
 gst_vaapi_video_allocator_new (GstVaapiDisplay * display,
-    const GstVideoInfo * vip, guint flags)
+    const GstVideoInfo * vip, guint surface_alloc_flags)
 {
   GstVaapiVideoAllocator *allocator;
 
@@ -752,7 +765,7 @@ gst_vaapi_video_allocator_new (GstVaapiDisplay * display,
 
   allocator_configure_surface_info (display, allocator);
   allocator->surface_pool = gst_vaapi_surface_pool_new_full (display,
-      &allocator->surface_info, flags);
+      &allocator->surface_info, surface_alloc_flags);
   if (!allocator->surface_pool)
     goto error_create_surface_pool;
 
@@ -837,6 +850,9 @@ gst_vaapi_dmabuf_memory_new (GstAllocator * allocator, GstVaapiVideoMeta * meta)
   gst_vaapi_video_meta_set_surface_proxy (meta, proxy);
   gst_vaapi_surface_proxy_unref (proxy);
 
+  /* Need dup because GstDmabufMemory creates the GstFdMemory with flag
+   * GST_FD_MEMORY_FLAG_NONE. So when being freed it calls close on the fd
+   * because GST_FD_MEMORY_FLAG_DONT_CLOSE is not set. */
   dmabuf_fd = gst_vaapi_buffer_proxy_get_handle (dmabuf_proxy);
   if (dmabuf_fd < 0 || (dmabuf_fd = dup (dmabuf_fd)) < 0)
     goto error_create_dmabuf_handle;
@@ -1047,3 +1063,42 @@ gst_allocator_set_vaapi_video_info (GstAllocator * allocator,
 
   return TRUE;
 }
+
+/**
+ * gst_allocator_get_vaapi_image_size:
+ * @allocator: a #GstAllocator instance.
+ * @size: (out) (optional): the VA image size created by @allocator.
+ *
+ * This function gets the size of the VA images instantiated by the
+ * @allocator.
+ *
+ * Returns: %TRUE if @allocator is VA valid
+ **/
+gboolean
+gst_allocator_get_vaapi_image_size (GstAllocator * allocator, guint * size)
+{
+  GstVaapiVideoAllocator *alloc;
+
+  g_return_val_if_fail (GST_IS_ALLOCATOR (allocator), FALSE);
+
+  if (g_strcmp0 (allocator->mem_type, GST_VAAPI_VIDEO_MEMORY_NAME))
+    return FALSE;
+  alloc = GST_VAAPI_VIDEO_ALLOCATOR_CAST (allocator);
+  if (alloc && size)
+    *size = GST_VIDEO_INFO_SIZE (&alloc->image_info);
+
+  return TRUE;
+}
+
+gboolean
+gst_vaapi_is_dmabuf_allocator (GstAllocator * allocator)
+{
+  GstStructure *st;
+
+  g_return_val_if_fail (GST_IS_ALLOCATOR (allocator), FALSE);
+
+  if (g_strcmp0 (allocator->mem_type, GST_ALLOCATOR_DMABUF) != 0)
+    return FALSE;
+  st = g_object_get_qdata (G_OBJECT (allocator), GST_VAAPI_VIDEO_INFO_QUARK);
+  return (st != NULL);
+}
